/*
	Fix Your Singer
	(c) 2008--2009 Stephane Magnenat <stephane at magnenat dot net>

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation version 3.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	The GPL can be found at <http://www.gnu.org/licenses/>.
*/


#include "PitchEditorWidget.h"
#include "PitchVector.h"
#include <QtGui>
#include <cmath>
#include <cassert>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include "ludecomposition.h"

#include <iostream>
using namespace std;

#define SEMI_TONE_FACTOR 1.05946309435930

/*template<class T>
void dumpMatrix(const ublas::matrix<T>& m)
{
	cout << "Matrix " << &m << "\n";
	for (size_t i = 0; i < m.size1(); ++ i)
	{
		for (size_t j = 0; j < m.size2(); ++ j)
			cout << m(i, j) << " ";
		cout << "\n";
	}
}*/

PitchEditorWidget::PitchEditorWidget(PitchVector* pitchData, QWidget* parent) :
	QWidget(parent),
	pitchData(pitchData),
	window(pitchData->bounds),
	scaleDisplay(SCALE_LATIN),
	selectionStart(0),
	selectionStop(0),
	selectionMode(SELECTION_NONE),
	editTool(TOOL_NONE),
	dragParamIndex(-1),
	panning(false)
{
	setAttribute(Qt::WA_OpaquePaintEvent, true);
}

void PitchEditorWidget::zoom(const QRectF& rect)
{
	window = rect; 
	repaint();
}

void PitchEditorWidget::selectToolConstant()
{
	setupCurveControlParams(1);
}

void PitchEditorWidget::selectToolLinear()
{
	setupCurveControlParams(2);
}

void PitchEditorWidget::selectToolSpline()
{
	setupCurveControlParams(4);
}

void PitchEditorWidget::setupCurveControlParams(unsigned dof)
{
	if (selectionMode == SELECTION_FULL)
	{
		int startIndex = pitchData->getIndexFromTime(selectionStart, true);
		int endIndex = pitchData->getIndexFromTime(selectionStop, false);
		assert(startIndex <= endIndex);
		unsigned count = endIndex - startIndex + 1;
		
		// we do not manage underparametrized systems, we do not like singular matrixes
		if (count < dof)
			return;
		
		// first we get parameters in vector a such that y(x) = a0 + a1 * x + a2 * x^2 + ... + an * x^n
		// minimizing the least square error
		
		curveControlParams.resize(dof);
		ublas::matrix<double> y(count, 1);
		ublas::matrix<double> X(count, dof);
		
		for (size_t i = 0; i < count; i++)
			y(i, 0) = pitchData->editedData[i+startIndex].y();
		
		for (size_t i = 0; i < count; i++)
			for (size_t j = 0; j < dof; j++)
				X(i, j) = pow(pitchData->editedData[i+startIndex].x(), (double)j);
		
		// solve (X^t X) a = X ^t y
		// == solve a = (X^t X)^-1 X^t y
		ublas::matrix<double> Xt(trans(X));
		ublas::matrix<double> XtX(prod(Xt, X));
		assert(XtX.size1() == XtX.size2());
		ublas::matrix<double> b(prod(Xt, y));
		
		// Solve a linear equation system a*x=b using inplace LU decomposition.
		//
		// Stores x in 'b' and overwrites 'a' (with a pivotted LUD).
		//
		// Matrix 'b' may have any (>0) number of columns but
		// must contain as many rows as 'a'.
		//
		// Possible return values:
		//  0=success
		//  1=singular matrix
		//  2=a.rows != b.rows
		int ret = LU_Solve(XtX, b);
		if (ret == 0)
		{
			curveControlParams = column(b, 0);
			
			if (dof == 1)
			{
				editTool = TOOL_CONSTANT;
			}
			else if (dof == 2)
			{
				editTool = TOOL_LINEAR;
			}
			else if (dof == 4)
			{
				editTool = TOOL_SPLINE;
			}
			else
				assert(false);
			
			std::cout << "curveControlParams\n" << curveControlParams << std::endl;
			repaint();
		}
	}
}

double PitchEditorWidget::evaluateCurveControl(double pos)
{
	return evaluateCurveControl(pos, curveControlParams);
}

double PitchEditorWidget::evaluateCurveControl(double pos, const ublas::vector<double> &params)
{
	double val = 0;
	for (size_t i = 0; i < params.size(); i++)
		val += params[i] * pow(pos, (double)i);
	return val;
}

double PitchEditorWidget::evaluateCurveControlDer(double pos, const ublas::vector<double> &params)
{
	double val = 0;
	for (size_t i = 1; i < params.size(); i++)
		val += params[i] * pow(pos, (double)(i-1)) * (double)i;
	return val;
}

void PitchEditorWidget::getSplineCP(ublas::vector<double> &cpX, ublas::vector<double> &cpY)
{
	// transform p[0] + p[1] x + p[2] x ^ 2 + p[3] x^3
	// to sp[0] (y1), sp[1] (dy1), sp[2] (dy2), sp[3] (y2) using selectionStart and selectionStop
	cpX[0] = selectionStart;
	cpX[1] = selectionStart + (selectionStop - selectionStart) / 3;
	cpX[2] = selectionStart + (2*(selectionStop - selectionStart)) / 3;
	cpX[3] = selectionStop;
	cpY[0] = evaluateCurveControl(selectionStart, curveControlParams);
	cpY[1] = cpY[0] + evaluateCurveControlDer(selectionStart, curveControlParams) * (cpX[1] - cpX[0]);
	cpY[3] = evaluateCurveControl(selectionStop, curveControlParams);
	cpY[2] = cpY[3] - evaluateCurveControlDer(selectionStop, curveControlParams) * (cpX[3] - cpX[2]);
}

void PitchEditorWidget::setSplineFromCP(const ublas::vector<double> &cpX, const ublas::vector<double> &cpY, ublas::vector<double> &np)
{
	double yds = (cpY[1] - cpY[0]) / (cpX[1] - cpX[0]);
	double ydd = (cpY[3] - cpY[2]) / (cpX[3] - cpX[2]);
	double xs = cpX[0];
	double ys = cpY[0];
	double xd = cpX[3];
	double yd = cpY[3];
	double A = (yds - ydd) / (2*xs - 2*xd);
	double B = (3*xd*xd - 3*xs*xs) / (2*xs-2*xd);
	double D = (-xs*xs+2*xs*xd-xd*xd);
	double E = (-2*xs*xs*xs+3*xd*xs*xs-xd*xd*xd);
	double F = (xs*yds-ys-xd*yds+yd);
	np[3] = - (F + A*D)/(B*D + E);
	np[2] = A + np[3] * B;
	np[1] = yds - 2*np[2]*xs - 3*np[3]*xs*xs;
	np[0] = ys - np[1]*xs - np[2]*xs*xs - np[3]*xs*xs*xs;
}

void PitchEditorWidget::toolReset()
{
	if (selectionMode == SELECTION_FULL)
	{
		int startIndex = pitchData->getIndexFromTime(selectionStart, true);
		int endIndex = pitchData->getIndexFromTime(selectionStop, false);
		for (int i = startIndex; i <= endIndex; i++)
			pitchData->editedData[i] = pitchData->originalData[i];
		editTool = TOOL_NONE;
		repaint();
	}
}

void PitchEditorWidget::setScaleLatin()
{
	scaleDisplay = SCALE_LATIN;
	repaint();
}

void PitchEditorWidget::setScaleAnglosaxon()
{
	scaleDisplay = SCALE_ANGLOSAXON;
	repaint();
}

void PitchEditorWidget::paintEvent ( QPaintEvent * event )
{
	QPainter painter(this);
	
	// draw background
	painter.fillRect(rect(), Qt::white);
	
	// draw pitch data
	QRect rect(-1,-1,3,3);
	QRect bigRect(-3,-3,7,7);
	int startIndex = pitchData->getIndexFromTime(window.left(), false);
	int endIndex = pitchData->getIndexFromTime(window.right(), true);
	if (startIndex >= 0)
	{
		assert(endIndex >= 0);
		
		// draw original pitch
		for (int i = startIndex; i <= endIndex; i++)
		{
			if (pitchData->originalData[i].y() == pitchData->editedData[i].y())
			{
				painter.fillRect(rect.translated(posToPixel(pitchData->originalData[i].x()), freqToPixel(pitchData->originalData[i].y())), Qt::darkBlue);
			}
			else
			{
				painter.fillRect(rect.translated(posToPixel(pitchData->originalData[i].x()), freqToPixel(pitchData->originalData[i].y())), Qt::lightGray);
				painter.fillRect(rect.translated(posToPixel(pitchData->editedData[i].x()), freqToPixel(pitchData->editedData[i].y())), Qt::blue);
			}
		}
		
		// draw selection
		if (selectionMode == SELECTION_ONE_POINT)
		{
			int startPos = posToPixel(selectionStart);
			painter.setPen(Qt::darkRed);
			painter.drawLine(startPos, 0, startPos, height());
		}
		else if (selectionMode == SELECTION_FULL)
		{
			int startPos = posToPixel(selectionStart);
			int stopPos = posToPixel(selectionStop);
			painter.fillRect(QRect(startPos, 0, stopPos-startPos, height()), QColor(200, 0, 0, 64));
			painter.setPen(Qt::darkRed);
			painter.drawLine(startPos, 0, startPos, height());
			painter.drawLine(stopPos, 0, stopPos, height());
		}
		
		// draw edit tools
		if (editTool == TOOL_CONSTANT)
		{
			double y = evaluateCurveControl(selectionStart);
			
			painter.drawLine(posToPixel(selectionStart), freqToPixel(y), posToPixel(selectionStop), freqToPixel(y));
			painter.fillRect(bigRect.translated(posToPixel(selectionStart), freqToPixel(y)), Qt::red);
			painter.drawRect(bigRect.translated(posToPixel(selectionStart), freqToPixel(y)));
			painter.fillRect(bigRect.translated(posToPixel(selectionStop), freqToPixel(y)), Qt::red);
			painter.drawRect(bigRect.translated(posToPixel(selectionStop), freqToPixel(y)));
		}
		else if (editTool == TOOL_LINEAR)
		{
			double y1 = evaluateCurveControl(selectionStart);
			double y2 = evaluateCurveControl(selectionStop);
			
			painter.drawLine(posToPixel(selectionStart), freqToPixel(y1), posToPixel(selectionStop), freqToPixel(y2));
			painter.fillRect(bigRect.translated(posToPixel(selectionStart), freqToPixel(y1)), Qt::red);
			painter.drawRect(bigRect.translated(posToPixel(selectionStart), freqToPixel(y1)));
			painter.fillRect(bigRect.translated(posToPixel(selectionStop), freqToPixel(y2)), Qt::red);
			painter.drawRect(bigRect.translated(posToPixel(selectionStop), freqToPixel(y2)));
		}
		else if (editTool == TOOL_SPLINE)
		{
			ublas::vector<double> cpX(4), cpY(4);
			getSplineCP(cpX, cpY);
			
			unsigned stepsCount = (posToPixel(selectionStop) - posToPixel(selectionStart)) / 4;
			stepsCount = std::max(stepsCount, (unsigned)3);
			for (unsigned i = 0; i < stepsCount; i++)
			{
				double dx0 = selectionStart + ((selectionStop - selectionStart)*(double)i) / (double)stepsCount;
				double dx1 = selectionStart + ((selectionStop - selectionStart)*(double)(i+1)) / (double)stepsCount;
				painter.drawLine(posToPixel(dx0), freqToPixel(evaluateCurveControl(dx0)), posToPixel(dx1), freqToPixel(evaluateCurveControl(dx1)));
			}
			
			painter.setPen(Qt::white);
			painter.drawLine(posToPixel(cpX[0]), freqToPixel(cpY[0]), posToPixel(cpX[1]), freqToPixel(cpY[1]));
			painter.drawLine(posToPixel(cpX[2]), freqToPixel(cpY[2]), posToPixel(cpX[3]), freqToPixel(cpY[3]));
			
			painter.setPen(Qt::darkRed);
			for (size_t i = 0; i < 4; i++)
			{
				painter.fillRect(bigRect.translated(posToPixel(cpX[i]), freqToPixel(cpY[i])), Qt::red);
				painter.drawRect(bigRect.translated(posToPixel(cpX[i]), freqToPixel(cpY[i])));
			}
		}
	}
	
	// draw scale, tempered
	painter.fillRect(QRect(0, 0, 80, height()), QColor(255, 255, 255, 224));
	painter.setPen(Qt::black);
	painter.drawLine(80, 0, 80, height());
	
	const char* notesNamesLatin[12] =
		{ "La", "La#", "Si", "Do", "Do#", "Re", "Re#", "Mi", "Fa", "Fa#", "Sol", "Sol#" };
	const char* notesNamesAnglosaxon[12] =
		{ "A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#" };
	const char** notesNames;
	if (scaleDisplay == SCALE_LATIN)
		notesNames = notesNamesLatin;
	else
		notesNames = notesNamesAnglosaxon;
	
	qreal factor = SEMI_TONE_FACTOR;
	qreal freq = 55;
	int noteIndex = 0;
	while ((freq < window.bottom()) && (freq < 3520+1))
	{
		if (freq > window.top())
		{
			int coord = freqToPixel(freq);
			
			if (noteIndex == 0)
			{
				painter.setPen(Qt::black);
				painter.drawLine(0, coord, width(), coord);
				painter.setPen(Qt::darkBlue);
				painter.drawText(QPointF(40, coord-1), QString("%0").arg(freq));
			}
			else
			{
				painter.setPen(Qt::gray);
				painter.drawLine(0, coord, width(), coord);
			}
			
			painter.setPen(Qt::black);
			painter.drawText(QPointF(5, coord-1), QString("%0").arg(notesNames[noteIndex]));
		}
		freq *= factor;
		noteIndex = (noteIndex + 1) % 12;
	}
	
	// draw time line
	painter.fillRect(QRect(0, 0, width(), 20), QColor(255, 255, 255, 224));
	painter.setPen(Qt::black);
	painter.drawLine(0, 20, width(), 20);
	double lastTimePos = -1;
	for (double timePos = round(window.left()); timePos < window.right(); timePos += 1)
	{
		int minutes = ((int)timePos) / 60;
		int seconds = ((int)timePos) % 60;
		int pixelPos = posToPixel(timePos);
		int lastPixelPos = posToPixel(lastTimePos);
		if (pixelPos > lastPixelPos + 50)
		{
			painter.setPen(Qt::black);
			painter.drawLine(pixelPos, 0, pixelPos, 20);
			painter.drawText(QPointF(pixelPos+3, 15), QString("%0:%1").arg(minutes).arg(seconds));
			lastTimePos = timePos;
			painter.setPen(Qt::gray);
			painter.drawLine(pixelPos, 21, pixelPos, height());
		}
	}
}

void PitchEditorWidget::mousePressEvent ( QMouseEvent * event )
{
	if (event->button() == Qt::LeftButton)
	{
		if (editTool != TOOL_NONE)
		{
			QRect bigRect(-3,-3,7,7);
			
			double y1 = evaluateCurveControl(selectionStart);
			QPoint p1(posToPixel(selectionStart), freqToPixel(y1));
			if (bigRect.translated(p1).contains(event->pos()))
			{
				dragParamIndex = 0;
				dragParamPos = event->pos();
				return;
			}
			
			double y2 = evaluateCurveControl(selectionStop);
			QPoint p2(posToPixel(selectionStop), freqToPixel(y2));
			if (bigRect.translated(p2).contains(event->pos()))
			{
				dragParamIndex = curveControlParams.size()-1;
				dragParamPos = event->pos();
				return;
			}
			
			if (editTool == TOOL_SPLINE)
			{
				ublas::vector<double> cpX(4), cpY(4);
				getSplineCP(cpX, cpY);
				for (size_t i = 0; i < 4; i++)
				{
					QPoint p(posToPixel(cpX[i]), freqToPixel(cpY[i]));
					if (bigRect.translated(p).contains(event->pos()))
					{
						dragParamIndex = i;
						dragParamPos = event->pos();
						return;
					}
				}
			}
		}
		
		if (selectionMode == SELECTION_NONE)
		{
			selectionStart = pixelToPos(event->x());
			selectionMode = SELECTION_ONE_POINT;
			repaint();
		}
		else if (selectionMode == SELECTION_ONE_POINT)
		{
			selectionStop = pixelToPos(event->x());
			if (selectionStop < selectionStart)
			{
				qreal tmp = selectionStart;
				selectionStart = selectionStop;
				selectionStop = tmp;
			}
			selectionMode = SELECTION_FULL;
			selectionStateChanged(true);
			repaint();
		}
		else
		{
			selectionStart = pixelToPos(event->x());
			selectionMode = SELECTION_ONE_POINT;
			editTool = TOOL_NONE;
			selectionStateChanged(false);
			repaint();
			
			/*selectionMode = SELECTION_NONE;
			editTool = TOOL_NONE;
			repaint();*/
		}
	}
	else if (event->button() == Qt::MidButton)
	{
		panning = true;
		panningPos = event->pos();
	}
	else if (event->button() == Qt::RightButton)
	{
		selectionMode = SELECTION_NONE;
		editTool = TOOL_NONE;
		selectionStateChanged(false);
		repaint();
	}
}

void PitchEditorWidget::mouseMoveEvent ( QMouseEvent * event )
{
	// TODO: unify panning and dragging
	if (panning)
	{
		QPoint diff = event->pos() - panningPos;
		panningPos = event->pos();
		
		if (!diff.isNull())
		{
			double dx = ((double)diff.x() * window.width()) / (double)width();
			window.translate(-dx, 0);
			
			// TODO: clean this copy pasted
			qreal minLinFreq = log(window.top()) / log(SEMI_TONE_FACTOR);
			qreal maxLinFreq = log(window.bottom()) / log(SEMI_TONE_FACTOR);
			qreal freqCount = maxLinFreq - minLinFreq;
			
			double dy = ((double)diff.y() * freqCount) / (double)height();
			window.setTop(window.top() * powf(SEMI_TONE_FACTOR, dy));
			window.setBottom(window.bottom() * powf(SEMI_TONE_FACTOR, dy));
			repaint();
		}
	}
	
	if (dragParamIndex >= 0)
	{
		QPoint diff = event->pos() - dragParamPos;
		dragParamPos = event->pos();
		
		if (!diff.isNull())
		{
			// TODO: clean this copy pasted
			qreal minLinFreq = log(window.top()) / log(SEMI_TONE_FACTOR);
			qreal maxLinFreq = log(window.bottom()) / log(SEMI_TONE_FACTOR);
			qreal freqCount = maxLinFreq - minLinFreq;
		
			double dy = ((double)diff.y() * freqCount) / (double)height();
			double df = powf(SEMI_TONE_FACTOR, dy);
			
			int startIndex = pitchData->getIndexFromTime(selectionStart, true);
			int endIndex = pitchData->getIndexFromTime(selectionStop, false);
			assert(startIndex <= endIndex);
			
			assert(dragParamIndex < curveControlParams.size());
			ublas::vector<double> np(curveControlParams.size());
			
			if (editTool == TOOL_CONSTANT)
			{
				np[0] = curveControlParams[0] / df;
			}
			else if (editTool == TOOL_LINEAR)
			{
				double y1 = evaluateCurveControl(selectionStart);
				double y2 = evaluateCurveControl(selectionStop);
				if (dragParamIndex == 0)
					y1 /= df;
				else
					y2 /= df;
				
				// p0 = (y2 - y1) / (x2 - x1)
				np[1] = (y2 - y1) / (selectionStop - selectionStart);
				np[0] = y1 - (np[1] * selectionStart);
			}
			else if (editTool == TOOL_SPLINE)
			{
				ublas::vector<double> cpX(4), cpY(4);
				getSplineCP(cpX, cpY);
				cpY[dragParamIndex] /= df;
				setSplineFromCP(cpX, cpY, np);
			}
			else
				assert(false);
			
			// apply diff
			for (int i = startIndex; i <= endIndex; i++)
			{
				double diff =  pitchData->editedData[i].y() - evaluateCurveControl(pitchData->editedData[i].x());
				pitchData->editedData[i].setY(diff + evaluateCurveControl(pitchData->editedData[i].x(), np));
			}
			
			curveControlParams = np;
			
			repaint();
		}
	}
}

void PitchEditorWidget::mouseReleaseEvent ( QMouseEvent * event )
{
	if (event->button() == Qt::LeftButton)
	{
		dragParamIndex = -1;
	}
	
	if (event->button() == Qt::MidButton)
	{
		panning = false;
	}
}

void PitchEditorWidget::wheelEvent ( QWheelEvent * event )
{
	if (event->modifiers() & Qt::ShiftModifier)
	{
		if (event->modifiers() & Qt::ControlModifier)
		{
			if (event->delta() > 0)
			{
				window.setTop(window.top() * sqrt(SEMI_TONE_FACTOR));
				window.setBottom(window.bottom() * sqrt(SEMI_TONE_FACTOR));
			}
			else
			{
				window.setTop(window.top() / sqrt(SEMI_TONE_FACTOR));
				window.setBottom(window.bottom() / sqrt(SEMI_TONE_FACTOR));
			}
		}
		else
		{
			if (event->delta() > 0)
			{
				window.setTop(window.top() * sqrt(SEMI_TONE_FACTOR));
				window.setBottom(window.bottom() / sqrt(SEMI_TONE_FACTOR));
			}
			else
			{
				window.setTop(window.top() / sqrt(SEMI_TONE_FACTOR));
				window.setBottom(window.bottom() * sqrt(SEMI_TONE_FACTOR));
			}
		}
	}
	else
	{
		if (event->modifiers() & Qt::ControlModifier)
		{
			if (event->delta() > 0)
			{
				window.setLeft(window.left() + window.width() / 20);
				window.setRight(window.right() + window.width() / 20);
			}
			else
			{
				window.setLeft(window.left() - window.width() / 20);
				window.setRight(window.right() - window.width() / 20);
			}
		}
		else
		{	
			if (event->delta() > 0)
			{
				window.setLeft(window.left() + window.width() / 20);
				window.setRight(window.right() - window.width() / 20);
			}
			else
			{
				window.setLeft(window.left() - window.width() / 20);
				window.setRight(window.right() + window.width() / 20);
			}
		}
	}
	repaint();
}

int PitchEditorWidget::freqToPixel(qreal freq)
{
	qreal minLinFreq = log(window.top()) / log(SEMI_TONE_FACTOR);
	qreal maxLinFreq = log(window.bottom()) / log(SEMI_TONE_FACTOR);
	qreal linFreq = log(freq) / log(SEMI_TONE_FACTOR);
	return height() - (int)(((linFreq - minLinFreq) * (qreal)height()) / (maxLinFreq - minLinFreq));
}

qreal PitchEditorWidget::pixelToFreq(int pos)
{
	return 0;
}

int PitchEditorWidget::posToPixel(qreal pos)
{
	return (int)(((pos - window.left()) * (qreal)width()) / window.width());
}

qreal PitchEditorWidget::pixelToPos(int pixel)
{
	return window.left() + ((qreal)pixel * window.width()) / (qreal)width();
}
